# vi: encoding=utf-8 ts=8 sts=4 sw=4 et

import re
import gzip
import time
import sys
from io import StringIO
import os
import fcntl
import string
import tempfile
import shutil

# TLD_Builder:
import gpg
import request
import util
import log

class B_Queue:
    def __init__(self, filename):
        self.name = filename
        self.requests = []
        self.fd = None

    def dump(self, fname):
        (fdno, tmpfname) = tempfile.mkstemp(dir=os.path.dirname(fname))
        f = os.fdopen(fdno, "w")
        self.requests.reverse()
        for r in self.requests:
            r.dump(f)
        self.requests.reverse()
        f.flush()
        os.fsync(f.fileno())
        f.close()
        os.chmod(tmpfname, 0o0644)
        os.rename(tmpfname, fname)

    def dump_html(self, fname):
        (fdno, tmpfname) = tempfile.mkstemp(dir=os.path.dirname(fname))
        f = os.fdopen(fdno, "w")
        f.write("""
<html>
    <head>
    <link rel="Shortcut Icon" href="//www.tld-linux.org/favicon.ico"/>
        <title>TLD builder queue</title>
        <link rel="stylesheet" type="text/css" charset="utf-8" media="all" href="style.css">
        <script src="//ajax.googleapis.com/ajax/libs/jquery/1.9.1/jquery.min.js"></script>
        <script src="//cdnjs.cloudflare.com/ajax/libs/jquery-timeago/1.4.3/jquery.timeago.min.js"></script>
        <script type="text/javascript" src="script.js"></script>
    </head>
<body>
<div id="header">
    <div id="requesters-filter">Filter by requesters:<br></div>
</div>
<div id="content">
"""
        )
        self.requests.reverse()
        for r in self.requests:
            r.dump_html(f)
        self.requests.reverse()
        f.write("</div></body></html>\n")
        f.flush()
        os.fsync(f.fileno())
        f.close()
        os.chmod(tmpfname, 0o0644)
        os.rename(tmpfname, fname)

    # read possibly compressed, signed queue
    def read_signed(self):
        if re.search(r"\.gz$", self.name):
            f = gzip.open(self.name)
        else:
            f = open(self.name)
        (signers, body) = gpg.verify_sig(f.read())
        self.signers = signers
        self.requests = request.parse_requests(body)

    def _open(self):
        if self.fd == None:
            if os.access(self.name, os.F_OK):
                self.fd = open(self.name, "r+")
            else:
                self.fd = open(self.name, "w+")

    def read(self):
        self._open()
        self.signers = []
        body = self.fd.read()
        if body.strip() == "":
            # empty file, don't choke
            self.requests = []
            return
        try:
            self.requests = request.parse_requests(body)
        except Exception as e:
            log.panic("error parsing %s: %s" % (self.name, e))
            pass

    def _write_to(self, f):
        f.write("<queue>\n")
        for r in self.requests:
            r.write_to(f)
        f.write("</queue>\n")

    def write(self):
        self._open()
        self.fd.seek(0)
        self.fd.truncate(0)
        self._write_to(self.fd)
        self.fd.flush()

    def lock(self, no_block):
        self._open()
        op = fcntl.LOCK_EX
        if no_block:
            op = op + fcntl.LOCK_NB
        try:
            fcntl.flock(self.fd, op)
            return 1
        except IOError:
            return 0

    def unlock(self):
        fcntl.flock(self.fd, fcntl.LOCK_UN)

    def write_signed(self, name):
        sio = StringIO()
        self._write_to(sio)
        sio.seek(0)
        sio.write(util.to_str(gpg.sign(sio.read())))
        sio.seek(0)
        (fdno, tmpname) = tempfile.mkstemp(dir=os.path.dirname(name))
        f = os.fdopen(fdno, "w")
        shutil.copyfileobj(sio, f)
        f.flush()
        os.fsync(f.fileno())
        f.close()
        if re.search(r"\.gz$", name):
            f_in = open(tmpname, 'rb')
            f_out = gzip.open(name, 'wb')
            shutil.copyfileobj(f_in, f_out)
            f_in.close();
            f_out.flush();
            os.fsync(f_out.fileno())
            f_out.close();
            os.remove(tmpname)
        else:
            os.rename(tmpname, name)
        os.chmod(name, 0o0644)

    def add(self, req):
        self.requests.append(req)

    def value(self):
        return self.requests
